@@ -8,7 +8,7 @@
     :license: BSD, see LICENSE for details.
 """
 
-from typing import Any, Dict, List, Tuple, Type, cast
+from typing import Any, Dict, List, Optional, Tuple, Type, cast
 
 from docutils import nodes
 from docutils.nodes import Element
@@ -22,10 +22,14 @@ from sphinx.locale import __
 from sphinx.transforms import SphinxTransform
 from sphinx.util import logging
 from sphinx.util.docutils import SphinxTranslator
-from sphinx.util.nodes import process_only_nodes
+from sphinx.util.nodes import find_pending_xref_condition, process_only_nodes
 
 logger = logging.getLogger(__name__)
 
+if False:
+    # For type annotation
+    from docutils.nodes import Node
+
 
 class SphinxPostTransform(SphinxTransform):
     """A base class of post-transforms.
@@ -34,8 +38,8 @@ class SphinxPostTransform(SphinxTransform):
     They do resolving references, convert images, special transformation for each output
     formats and so on.  This class helps to implement these post transforms.
     """
-    builders = ()   # type: Tuple[str, ...]
-    formats = ()    # type: Tuple[str, ...]
+    builders: Tuple[str, ...] = ()
+    formats: Tuple[str, ...] = ()
 
     def apply(self, **kwargs: Any) -> None:
         if self.is_supported():
@@ -97,14 +101,27 @@ class ReferencesResolver(SphinxPostTransform):
                     if newnode is None:
                         self.warn_missing_reference(refdoc, typ, target, node, domain)
             except NoUri:
-                newnode = contnode
-            node.replace_self(newnode or contnode)
+                newnode = None
+
+            if newnode:
+                newnodes: List[Node] = [newnode]
+            else:
+                newnodes = [contnode]
+                if newnode is None and isinstance(node[0], addnodes.pending_xref_condition):
+                    matched = find_pending_xref_condition(node, "*")
+                    if matched:
+                        newnodes = matched.children
+                    else:
+                        logger.warning(__('Could not determine the fallback text for the '
+                                          'cross-reference. Might be a bug.'), location=node)
+
+            node.replace_self(newnodes)
 
     def resolve_anyref(self, refdoc: str, node: pending_xref, contnode: Element) -> Element:
         """Resolve reference generated by the "any" role."""
         stddomain = self.env.get_domain('std')
         target = node['reftarget']
-        results = []  # type: List[Tuple[str, Element]]
+        results: List[Tuple[str, Element]] = []
         # first, try resolving as :doc:
         doc_ref = stddomain.resolve_xref(self.env, refdoc, self.app.builder,
                                          'doc', target, node, contnode)
@@ -125,7 +142,7 @@ class ReferencesResolver(SphinxPostTransform):
                 for role in domain.roles:
                     res = domain.resolve_xref(self.env, refdoc, self.app.builder,
                                               role, target, node, contnode)
-                    if res and isinstance(res[0], nodes.Element):
+                    if res and len(res) > 0 and isinstance(res[0], nodes.Element):
                         results.append(('%s:%s' % (domain.name, role), res))
         # now, see how many matches we got...
         if not results:
@@ -150,7 +167,7 @@ class ReferencesResolver(SphinxPostTransform):
         return newnode
 
     def warn_missing_reference(self, refdoc: str, typ: str, target: str,
-                               node: pending_xref, domain: Domain) -> None:
+                               node: pending_xref, domain: Optional[Domain]) -> None:
         warn = node.get('refwarn')
         if self.config.nitpicky:
             warn = True
@@ -168,14 +185,13 @@ class ReferencesResolver(SphinxPostTransform):
         if self.app.emit_firstresult('warn-missing-reference', domain, node):
             return
         elif domain and typ in domain.dangling_warnings:
-            msg = domain.dangling_warnings[typ]
+            msg = domain.dangling_warnings[typ] % {'target': target}
         elif node.get('refdomain', 'std') not in ('', 'std'):
-            msg = (__('%s:%s reference target not found: %%(target)s') %
-                   (node['refdomain'], typ))
+            msg = (__('%s:%s reference target not found: %s') %
+                   (node['refdomain'], typ, target))
         else:
-            msg = __('%r reference target not found: %%(target)s') % typ
-        logger.warning(msg % {'target': target},
-                       location=node, type='ref', subtype=typ)
+            msg = __('%r reference target not found: %s') % (typ, target)
+        logger.warning(msg, location=node, type='ref', subtype=typ)
 
 
 class OnlyNodeTransform(SphinxPostTransform):
@@ -190,13 +206,9 @@ class OnlyNodeTransform(SphinxPostTransform):
 
 
 class SigElementFallbackTransform(SphinxPostTransform):
-    """Fallback desc_sig_element nodes to inline if translator does not supported them."""
+    """Fallback various desc_* nodes to inline if translator does not supported them."""
     default_priority = 200
 
-    SIG_ELEMENTS = [addnodes.desc_sig_name,
-                    addnodes.desc_sig_operator,
-                    addnodes.desc_sig_punctuation]
-
     def run(self, **kwargs: Any) -> None:
         def has_visitor(translator: Type[nodes.NodeVisitor], node: Type[Element]) -> bool:
             return hasattr(translator, "visit_%s" % node.__name__)
@@ -206,24 +218,35 @@ class SigElementFallbackTransform(SphinxPostTransform):
             # subclass of SphinxTranslator supports desc_sig_element nodes automatically.
             return
 
-        if all(has_visitor(translator, node) for node in self.SIG_ELEMENTS):
-            # the translator supports all desc_sig_element nodes
-            return
-        else:
-            self.fallback()
+        # for the leaf elements (desc_sig_element), the translator should support _all_
+        if not all(has_visitor(translator, node) for node in addnodes.SIG_ELEMENTS):
+            self.fallback(addnodes.desc_sig_element)
 
-    def fallback(self) -> None:
-        for node in self.document.traverse(addnodes.desc_sig_element):
+        if not has_visitor(translator, addnodes.desc_inline):
+            self.fallback(addnodes.desc_inline)
+
+    def fallback(self, nodeType: Any) -> None:
+        for node in self.document.traverse(nodeType):
             newnode = nodes.inline()
             newnode.update_all_atts(node)
             newnode.extend(node)
             node.replace_self(newnode)
 
 
+class PropagateDescDomain(SphinxPostTransform):
+    """Add the domain name of the parent node as a class in each desc_signature node."""
+    default_priority = 200
+
+    def run(self, **kwargs: Any) -> None:
+        for node in self.document.traverse(addnodes.desc_signature):
+            node['classes'].append(node.parent['domain'])
+
+
 def setup(app: Sphinx) -> Dict[str, Any]:
     app.add_post_transform(ReferencesResolver)
     app.add_post_transform(OnlyNodeTransform)
     app.add_post_transform(SigElementFallbackTransform)
+    app.add_post_transform(PropagateDescDomain)
 
     return {
         'version': 'builtin',
